"""Prediction set construction methods"""
from pyhealth.calib.predictionset.favmac import FavMac
from pyhealth.calib.predictionset.label import LABEL
from pyhealth.calib.predictionset.scrib import SCRIB

__all__ = ['LABEL', 'SCRIB', 'FavMac']
